/*
 * Main.vala
 *
 * Copyright 2015 Tony George <teejee2008@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 *
 *
 */

using GLib;
using Gtk;
using Gee;
using Json;

using TeeJee.Logging;
using TeeJee.FileSystem;
using TeeJee.JSON;
using TeeJee.ProcessManagement;
using TeeJee.GtkHelper;
using TeeJee.System;
using TeeJee.Misc;

public Main App;
public const string AppName = "Crunchy Archive Manager";
public const string AppShortName = "crunchy";
public const string AppVersion = "16.5.2";
public const string AppAuthor = "Tony George";
public const string AppAuthorEmail = "teejeetech@gmail.com";

const string GETTEXT_PACKAGE = "";
const string LOCALE_DIR = "/usr/share/locale";

extern void exit(int exit_code);

public class Main : GLib.Object {
	public bool gui_mode = false;
	public string user_login = "";
	public string user_home = "";
	public int user_uid = -1;
	public bool add_context_menu = true;
	public bool add_context_submenu = false;
	public bool associate_archives = true;
	public string last_input_dir = "";
	public string last_output_dir = "";
	public bool first_run = true;
		
	public string temp_dir = "";
	public string current_dir = "";
	public string share_dir = "/usr/share";
	public string app_conf_path = "";

	public ArchiveFile archive;
	public Archiver archiver;
	public DesktopApp crunchy_app;
	public Gee.ArrayList<MimeType> mimetype_list;
	public Json.Object default_config;
	
	public AppMode app_mode = AppMode.OPEN;
	public Gee.ArrayList<string> arg_files;
	public string arg_outpath = "";
	public bool arg_same_path = false;
	public bool arg_prompt_outpath = false;
	public bool use_large_icons_nav_pane = false;
	public bool use_large_icons_file_pane = false;
	public bool hide_nav_pane = true;
	
	public string status_line = "";
	public int64 progress_count;
	public int64 progress_total;
	
	public string[] supported_formats_open;

	public static string[] extensions_tar = {
		".tar"
	};
	
	public static string[] extensions_tar_compressed = {
		".tar.gz", ".tgz",
		".tar.bzip2",".tar.bz2", ".tbz", ".tbz2", ".tb2",
		".tar.lzma", ".tar.lz", ".tlz",
		".tar.lzo",
		".tar.xz", ".txz"
	};

	public static string[] extensions_tar_packed = {
		".tar.7z",
		".tar.zip",
		".deb"
	};

	public static string[] extensions_7z_unpack = {
		".7z" , ".lzma",
		".bz2", ".bzip2",
		".gz" , ".gzip",
		".zip", ".jar", ".war", ".ear",
		".rar", ".cab", ".arj", ".z", ".taz", ".cpio",
		".rpm", ".deb",
		".lzh", ".lha",
		".chm", ".chw", ".hxs",
		".iso", ".dmg", ".dar", ".xar", ".hfs", ".ntfs", ".fat", ".vhd", ".mbr",
		".wim", ".swm", ".squashfs", ".cramfs", ".scap"
	};

	public static string[] extensions_single_file = {
		".bz2", ".gz", ".xz", ".lzo"
	};
	
	public static string[] formats_single_file = {
		"bz2", "gz", "xz", "lzo"
	}; //7z,zip,tar support multiple files
				
	public Main(string[] args, bool _gui_mode) {

		gui_mode = _gui_mode;

		arg_files = new Gee.ArrayList<string>();
		
		//config file
		string home = Environment.get_home_dir();
		app_conf_path = home + "/.config/crunchy.json";

		//load settings
		load_app_config();

		// create default objects
		archive = new ArchiveFile();
		archiver = new Archiver();
		archiver.archive = archive;
		
		supported_formats_open = {
			".tar",
			".tar.gz", ".tgz",
			".tar.bzip2", ".tar.bz2", ".tbz", ".tbz2", ".tb2",
			".tar.lzma", ".tar.lz", ".tlz",
			".tar.xz", ".txz",
			".tar.7z",
			".tar.zip",
			".7z", ".lzma",
			".bz2", ".bzip2",
			".gz", ".gzip",
			".zip", ".rar", ".cab", ".arj", ".z", ".taz", ".cpio",
			".rpm", ".deb",
			".lzh", ".lha",
			".chm", ".chw", ".hxs",
			".iso", ".dmg", ".xar", ".hfs", ".ntfs", ".fat", ".vhd", ".mbr",
			".wim", ".swm", ".squashfs", ".cramfs", ".scap"
		};

		//check dependencies
		string message;
		if (!check_dependencies(out message)) {
			if (gui_mode) {
				string title = _("Missing Dependencies");
				gtk_messagebox(title, message, null, true);
			}
			exit(1);
		}

		if (Archiver.7zip_version < 0){
			string title = _("Missing Dependencies");
			message = _("7-Zip version is unsupported") + " (v%s)\n".printf(Archiver.7zip_version_name);
			message += _("Please report this by sending an email to teejeetech@gmail.com");
			gtk_messagebox(title, message, null, true);
			exit (1);
		}

		//initialize current_dir as current directory for CLI mode
		if (!gui_mode) {
			current_dir = Environment.get_current_dir();
		}

		try {
			//create temp dir
			temp_dir = get_temp_file_path();

			var f = File.new_for_path(temp_dir);
			if (f.query_exists()) {
				Posix.system("rm -rf %s".printf(temp_dir));
			}
			f.make_directory_with_parents();
		}
		catch (Error e) {
			log_error (e.message);
		}

		//get user info
		user_login = get_user_login();
		user_home = "/home/" + user_login;
		user_uid = get_user_id(user_login);

		MimeType.query_mimetypes();
		DesktopApp.query_apps();
		MimeApp.query_mimeapps(user_home);
		
		foreach(var app in DesktopApp.applist.values){
			if (app.desktop_file_name == "crunchy.desktop"){
				crunchy_app = app;
				break;
			}
		}

		load_mimetype_list();
	}

	private void load_mimetype_list(){
		
		var mime_list = new Gee.ArrayList<string>();
		string mimelist = "/usr/share/%s/mimetypes".printf(AppShortName);
		if (file_exists(mimelist)){
			foreach(string line in file_read(mimelist).split("\n")){
				mime_list.add(line.strip());
			}
		}
		
		var list = new Gee.ArrayList<MimeType>();
		foreach(string key in MimeType.mimetypes.keys) {
			if (mime_list.contains(key)){
				var mime = MimeType.mimetypes[key];
				//mime.is_selected = true; // let user select explicitly
				list.add(mime);
			}
		}

		list.sort((a, b) => {
			return strcmp(a.comment,b.comment);
		});

		mimetype_list = list;
	}

	public bool check_dependencies(out string msg) {
		msg = "";

		string[] dependencies = { "grep", "find", "7z", "tar", "gzip", "xdg-mime" };

		string path;
		foreach(string cmd_tool in dependencies) {
			path = get_cmd_path (cmd_tool);
			if ((path == null) || (path.length == 0)) {
				msg += " * " + cmd_tool + "\n";
			}
		}

		if (msg.length > 0) {
			msg = _("Commands listed below are not available on this system") + ":\n\n" + msg + "\n";
			msg += _("Please install required packages and try running again");
			log_msg(msg);
			return false;
		}
		else {
			return true;
		}
	}

	public void update_context_menu(){
		update_context_menu_thunar();
		update_context_menu_nautilus();
		update_context_menu_nemo();
	}

	private Gee.ArrayList<ThunarMenuEntry> get_thunar_entries(){
		var list = new Gee.ArrayList<ThunarMenuEntry>();

		var action = new ThunarMenuEntry();
		action.name = "Compress...";
		action.icon = "crunchy";
		action.id = "crunchy-compress";
		action.command = "crunchy-gtk --compress %F";
		action.description = "Add selected files and folders to a new archive";
		action.patterns = "*";
		action.show_for_directories = true;
		action.show_for_audio_files = true;
		action.show_for_video_files = true;
		action.show_for_image_files = true;
		action.show_for_text_files = true;
		action.show_for_other_files = true;
		list.add(action);

		action = new ThunarMenuEntry();
		action.name = "Extract to <folder>";
		action.icon = "crunchy";
		action.id = "crunchy-extract-here";
		action.command = "crunchy-gtk --extract-same %F";
		action.description = "Extract files to subfolder <archive-name>";
		action.patterns = "*";
		action.show_for_directories = false;
		action.show_for_audio_files = false;
		action.show_for_video_files = false;
		action.show_for_image_files = false;
		action.show_for_text_files = false;
		action.show_for_other_files = true;
		list.add(action);

		action = new ThunarMenuEntry();
		action.name = "Extract To...";
		action.icon = "crunchy";
		action.id = "crunchy-extract-to";
		action.command = "crunchy-gtk --extract-ask %F";
		action.description = "Extract files to specified location";
		action.patterns = "*";
		action.show_for_directories = false;
		action.show_for_audio_files = false;
		action.show_for_video_files = false;
		action.show_for_image_files = false;
		action.show_for_text_files = false;
		action.show_for_other_files = true;
		list.add(action);
		
		return list;
	}
	
	private void update_context_menu_thunar(){
		var cmd_name = "thunar";
		var path = get_cmd_path (cmd_name);
		if ((path == null) || (path.length == 0)) {
			return; // not installed
		}

		ThunarMenuEntry.query_actions(user_home);
		
		if (add_context_menu){
			foreach(var entry in get_thunar_entries()){
				entry.add();
			}
		}
		else{
			foreach(var entry in get_thunar_entries()){
				entry.remove();
			}
		}
	}

	private void update_context_menu_nautilus(){
		var cmd_name = "nautilus";
		var path = get_cmd_path (cmd_name);
		if ((path == null) || (path.length == 0)) {
			return; // not installed
		}

		string[] file_names_sub = { "crunchy" };
		string[] file_names = { "crunchy-compress","crunchy-extract-here","crunchy-extract-to" };
			
		if (add_context_menu){
			
			string[] list_add = file_names;
			string[] list_remove = file_names_sub;
			
			if (add_context_submenu){
				list_add = file_names_sub;
				list_remove = file_names;
			}

			// remove
			foreach(string file_name in list_remove){
				var file_py = "%s/.local/share/nautilus-python/extensions/%s.py".printf(
					user_home, file_name);
					
				var file_pyc = "%s/.local/share/nautilus-python/extensions/%s.pyc".printf(
					user_home, file_name);

				if (file_exists(file_py)){
					log_debug("Removed Nautilus action: %s".printf(file_name));
					file_delete(file_py);
				}
				if (file_exists(file_pyc)){
					file_delete(file_pyc);
				}
			}

			// add
			foreach(string file_name in list_add){
				var src_file = "/usr/share/crunchy/actions/%s.py".printf(file_name);
				var dst_file = "%s/.local/share/nautilus-python/extensions/%s.py".printf(
					user_home, file_name);

				if (file_exists(src_file) && !file_exists(dst_file)){
					dir_create(file_parent(dst_file));
					file_copy(src_file,dst_file);
					log_debug("Added Nautilus action: %s".printf(file_name));
				}
			}
		}
		else{
			foreach(string file_name in new string[] {
				"crunchy","crunchy-compress","crunchy-extract-here","crunchy-extract-to" }){

				var file_py = "%s/.local/share/nautilus-python/extensions/%s.py".printf(
					user_home, file_name);
					
				var file_pyc = "%s/.local/share/nautilus-python/extensions/%s.pyc".printf(
					user_home, file_name);

				if (file_exists(file_py)){
					log_debug("Removed Nautilus action: %s".printf(file_name));
					file_delete(file_py);
				}
				if (file_exists(file_pyc)){
					file_delete(file_pyc);
				}
			}
		}
	}

	private void update_context_menu_nemo(){
		var cmd_name = "nemo";
		var path = get_cmd_path (cmd_name);
		if ((path == null) || (path.length == 0)) {
			return; // not installed
		}

		if (add_context_menu){
			foreach(string file_name in new string[] {
				"crunchy-compress","crunchy-extract-here","crunchy-extract-to" }){

				var src_file = "/usr/share/crunchy/actions/%s.nemo_action".printf(file_name);
				var dst_file = "/home/%s/.local/share/nemo/actions/%s.nemo_action".printf(
					user_login, file_name);

				if (file_exists(src_file) && !file_exists(dst_file)){
					dir_create(file_parent(dst_file));
					file_copy(src_file,dst_file);
					log_debug("Added Nemo action: %s".printf(file_name));
				}
			}
		}
		else{
			foreach(string file_name in new string[] {
				"crunchy-compress","crunchy-extract-here","crunchy-extract-to" }){

				var file_nemo = "/home/%s/.local/share/nemo/actions/%s.nemo_action".printf(
					user_login, file_name);
					
				if (file_exists(file_nemo)){
					file_delete(file_nemo);
					log_debug("Removed Nemo action: %s".printf(file_name));
				}
			}
		}
	}

	/* Common */

	public string create_log_dir() {
		string log_dir = "%s/.local/logs/%s".printf(user_home, AppShortName);
		dir_create(log_dir);
		return log_dir;
	}

	public void save_app_config() {

		var config = new Json.Object();

		if (archiver != null){
			config = archiver.to_json();
		}

		config.set_string_member("first_run", first_run.to_string());
		config.set_string_member("last_input_dir", last_input_dir);
		config.set_string_member("last_output_dir", last_output_dir);
		config.set_string_member("add_context_menu", add_context_menu.to_string());
		config.set_string_member("add_context_submenu", add_context_submenu.to_string());
		config.set_string_member("use_large_icons_nav_pane", use_large_icons_nav_pane.to_string());
		config.set_string_member("use_large_icons_file_pane", use_large_icons_file_pane.to_string());
		config.set_string_member("hide_nav_pane", hide_nav_pane.to_string());
		
		var json = new Json.Generator();
		json.pretty = true;
		json.indent = 2;
		var node = new Json.Node(NodeType.OBJECT);
		node.set_object(config);
		json.set_root(node);

		try {
			json.to_file(this.app_conf_path);
		} catch (Error e) {
			log_error (e.message);
		}

		log_msg("\n" + _("App config saved") + ": '%s'".printf(app_conf_path));
	}

	public void load_app_config() {
		var f = File.new_for_path(app_conf_path);
		if (!f.query_exists()) {
			return;
		}

		var parser = new Json.Parser();
		try {
			parser.load_from_file(this.app_conf_path);
		}
		catch (Error e) {
			log_error (e.message);
		}

		var node = parser.get_root();
		var config = node.get_object();

		default_config = config;

		first_run = json_get_bool(config, "first_run", true);
		last_input_dir = json_get_string(config, "last_input_dir", "");
		last_output_dir = json_get_string(config, "last_output_dir", "");
		add_context_menu = json_get_bool(config, "add_context_menu", true);
		add_context_submenu = json_get_bool(config, "add_context_submenu", false);
		use_large_icons_nav_pane = json_get_bool(config, "use_large_icons_nav_pane", false);
		use_large_icons_file_pane = json_get_bool(config, "use_large_icons_file_pane", false);
		hide_nav_pane = json_get_bool(config, "hide_nav_pane", false);

		log_msg(_("App config loaded") + ": '%s'".printf(this.app_conf_path));
	}

	public void exit_app() {
		save_app_config();
		clear_tmp();
		log_msg(_("Exiting Application"));
	}

	/* Core */

	public void compress(bool wait = false) {
		archiver.compress(archive, wait);
	}

	public void extract(bool wait = false) {
		archiver.extract(archive, wait);
	}

	public void test(bool wait = false) {
		// create a unique temporary extraction directory
		archiver.extraction_path = "%s/%s".printf(App.temp_dir,random_string());
		dir_create(archiver.extraction_path);

		// begin
		archiver.test(archive, wait);
	}
	
	public void open(bool wait = false) {
		archiver.open(archive, wait);
	}

	public void open_info(bool wait = false) {
		archiver.archive = this.archive;
		archiver.open_info(archive, wait);
	}

	public void set_extraction_path_from_args(){
		
		// set extraction path
		
		if (arg_same_path){
			// select a subfolder in source path for extraction
			archiver.extraction_path = "%s/%s".printf(
				file_parent(archive.file_path),
				file_title(archive.file_path));

			/* since user has not specified the directory we need to
			 * make sure that files are not overwritten accidentally
			 * in existing directories */
			 
			// create a unique extraction directory
			int count = 0;
			string outpath = archiver.extraction_path;
			while (dir_exists(outpath)||file_exists(outpath)){
				log_debug("dir_exists: %s".printf(outpath));
				outpath = "%s (%d)".printf(archiver.extraction_path, ++count);
			}
			log_debug("create_dir: %s".printf(outpath));
			archiver.extraction_path = outpath;
		}
		else if (arg_prompt_outpath){
			// do nothing
		}
		else {
			// set path specified on command line
			archiver.extraction_path = arg_outpath;
		}
	}

	public void merge_nested_extraction_directories(){
		if (arg_same_path && dir_exists(archiver.extraction_path)){
			try {
				File file = File.parse_name(archiver.extraction_path);
				FileInfo info;
				var list = new Gee.ArrayList<string>();
				var enumerator = file.enumerate_children ("%s".printf(FileAttribute.STANDARD_NAME), 0);
				while ((info = enumerator.next_file()) != null) {
					string child_path = GLib.Path.build_filename(
						archiver.extraction_path,
						info.get_name());
						
					list.add(child_path);
				}

				if ((list.size == 1) && (dir_exists(list[0]))){
					// move the nested folder one level up
					var src = list[0];
					var dst = archiver.extraction_path;
					var dst_tmp = GLib.Path.build_filename(file_parent(dst), random_string());
					if (dir_exists(src)){
						file_move(src, dst_tmp);
					}
					file_delete(dst);
					file_move(dst_tmp, dst);
				}
			}
			catch (Error e) {
				log_error (e.message);
			}
		}
	}
	
	public ArchiveFile new_archive() {
		archive = new ArchiveFile();
		archiver.action = ArchiveAction.CREATE;
		return archive;
	}

	public string get_random_password(){
		string stdout, stderr;
		exec_script_sync("head /dev/urandom | tr -dc 'a-zA-Z0-9-_!@#$%^&*()_+{}|:<>?=' | head -c 20", out stdout, out stderr, true);
		return stdout;
	}
}

public enum AppMode {
	NEW,
	CREATE,
	OPEN,
	TEST,
	EXTRACT
}

