/*
 * AptikCompGtk.vala
 *
 * Copyright 2015 Tony George <teejee2008@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 *
 *
 */

using GLib;
using Gtk;
using Gee;
using Json;

using TeeJee.Logging;
using TeeJee.FileSystem;
using TeeJee.JSON;
using TeeJee.ProcessManagement;
using TeeJee.GtkHelper;
using TeeJee.System;
using TeeJee.Misc;

public class CrunchyGtk : GLib.Object {

	public static int main (string[] args) {
		set_locale();

		Gtk.init(ref args);

		init_tmp(AppShortName);

		LOG_TIMESTAMP = false;

		App = new Main(args, true);
		parse_arguments(args);

		switch (App.app_mode) {
		case AppMode.CREATE:
			/* Only one archive will be created in one run */

			// create task
			App.archive = new ArchiveFile();
			App.archiver.action = ArchiveAction.CREATE;
			//App.archive.add_items(App.arg_files);
			App.archiver.load_from_json(App.default_config);

			// show compression settings window
			var window = new CreateArchiveWindow();
			window.destroy.connect(Gtk.main_quit);
			//window.show_all();
			break;
			
		case AppMode.EXTRACT:
		case AppMode.TEST:
			/* Only one archive will be extracted in one run */
			 
			// create task
			App.archive = new ArchiveFile(App.arg_files[0]);
			/* extraction path will be set by App.extract() and App.test() */

			if (App.app_mode == AppMode.EXTRACT){
				App.archiver.action = ArchiveAction.EXTRACT;
				App.set_extraction_path_from_args();
			}
			else{
				App.archiver.action = ArchiveAction.TEST;
			}

			var window = new ProgressWindow();
			window.destroy.connect(Gtk.main_quit);
			//window.show_all();
			break;
			
		//case AppMode.NEW:
		case AppMode.OPEN:
			/* There may be multiple files to be opened in one run.
			 * App.archive will be created in ViewArchiveWindow
			 * */ 
			
			var window = new ViewArchiveWindow();
			window.destroy.connect(Gtk.main_quit);
			//window.show_all();
			break;
			
		default:
			var window = new ViewArchiveWindow();
			window.destroy.connect(Gtk.main_quit);
			//window.show_all();
			break;
		}

		//start event loop
		Gtk.main();

		//if (App.archiver.dis_out != null){
		//	log_msg(App.archiver.dis_out.is_closed().to_string());
		//}
		
		App.exit_app();

		return 0;
	}

	private static void set_locale() {
		Intl.setlocale(GLib.LocaleCategory.MESSAGES, AppShortName);
		Intl.textdomain(GETTEXT_PACKAGE);
		Intl.bind_textdomain_codeset(GETTEXT_PACKAGE, "utf-8");
		Intl.bindtextdomain(GETTEXT_PACKAGE, LOCALE_DIR);
	}

	public static bool parse_arguments(string[] args) {

		string txt = "";
		foreach(string arg in args){
			txt += " %s".printf(arg);
		}

		//Gtk.Clipboard clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD);
		//clipboard.set_text (txt, -1);
		//gtk_messagebox("",txt,null,false);
		
		//parse options
		for (int k = 1; k < args.length; k++) // Oth arg is app path
		{
			switch (args[k].down()) {
			case "--create":
			case "--compress":
				App.app_mode = AppMode.CREATE;
				break;

			case "--test":
			case "--extract":
			case "--extract-same":
			case "--extract-ask":
				if (args[k].down() == "--test"){
					App.app_mode = AppMode.TEST;
				}
				else{
					App.app_mode = AppMode.EXTRACT;
				}

				if (args[k].down() == "--extract-same"){
					App.arg_same_path = true;
				}
				
				if (args[k].down() == "--extract-ask"){
					App.arg_prompt_outpath = true;
				}
				
				k++;
				if (file_exists(args[k])){
					App.arg_files.add(args[k]);
				}
				else{
					log_error(_("File not found") + ": %s".printf(args[k]));
					exit(1);
				}
				break;

			case "--output-path":
				k++;
				App.arg_outpath = args[k];
				if(!dir_exists(App.arg_outpath)){
					dir_create(App.arg_outpath);
				}
				break;
				
			//case "--new":
				//App.app_mode = AppMode.NEW;
				//break;

			case "--open":
				App.app_mode = AppMode.OPEN;
				App.archiver.action = ArchiveAction.LIST;
				k++;
				if (file_exists(args[k])){
					App.arg_files.add(args[k]);
				}
				else{
					log_error(_("File not found") + ": %s".printf(args[k]));
					exit(1);
				}
				break;

			case "--debug":
				LOG_DEBUG = true;
				break;

			case "--help":
			case "--h":
			case "-h":
				log_msg(help_message());
				exit(0);
				return true;

			default:
				// check if argument is a file
				if (!file_exists(args[k]) && !dir_exists(args[k])) {
					log_error(_("File not found") + ": %s".printf(args[k]));
					log_msg(help_message());
					exit(1);
					return false;
				}

				// add to list
				App.arg_files.add(args[k]);

				break;
			}
		}

		return true;
	}

	public static string help_message() {
		string msg = "\n" + AppName + " v" + AppVersion + " by Tony George (teejee2008@gmail.com)" + "\n";
		msg += "\n";
		msg += _("Syntax") + ": %s [options] <files>\n".printf(AppShortName);
		msg += "\n";
		msg += _("Options") + ":\n";
		msg += "\n";
		msg += " --create <files>  Displays 'Create Archive' window for selected files\n";
		msg += "\n";
		//msg += " --new <files>     Displays selected files in main window for creating new archive\n";
		//msg += "\n";
		msg += " --extract <archive>      Extract archive to specified path\n";
		msg += " --extract-same <archive> Extract to a subfolder in same directory\n";
		msg += " --extract-ask <archive>  Prompt for output directory and extract\n";
		msg += " --output-path <path>     Extract to specified path\n";
		msg += "                          (ignored if: '--extract-same', '--extract-ask')\n";
		msg += "\n";
		msg += _("Examples") + ":\n";
		msg += "\n";
		msg += "> %s --create <files>\n".printf(AppShortName);
		msg += "Displays compression settings dialog\n";
		msg += "\n";
		msg += "> %s --extract <archive-path> --output-path <directory-path>\n".printf(AppShortName);
		msg += "Extracts archive contents to specified path\n";
		msg += "\n";
		msg += "> %s --extract <archive-path>\n".printf(AppShortName);
		msg += "Extracts archive to subfolder in the same path\n";
		msg += "\n";
		msg += "> %s --extract-ask <archive-path>\n".printf(AppShortName);
		msg += "Prompts for output path and extracts the archive\n";
		msg += "\n";
		msg += "> %s --extract-same <archive-path>\n".printf(AppShortName);
		msg += "Extracts archive to subfolder in the same path\n";
		msg += "\n";
		return msg;
	}
}



